"""Initialize and run the application."""
import sys
from .application import main


def run_main():
    """Run the main application."""
    rc = main()
    sys.exit(rc)


if __name__ == "__main__":
    run_main()



"""Unit tests for application.py."""
import argparse
import pytest
import unittest.mock
import pyrebar.application

from importlib.metadata import EntryPoint
from pyrebar.plugins import PluginModule, ProcessedPlugins, Plugins


@pytest.fixture
def full_plugin() -> PluginModule:
    """Provide a plugin fixture.

    Returns:
        PluginModule: a mock module
    """
    return PluginModule(
        helpstr="help string",
        command="example_app",
        conf=unittest.mock.MagicMock(),
        func=unittest.mock.MagicMock(),
        aliases=["ex", "app"],
        logger_name="example.pyapp",
    )


@pytest.fixture
def processed_plugins() -> ProcessedPlugins:
    """Bootstrap the plugins."""
    return ProcessedPlugins(
        pre_init=(
            EntryPoint(
                name="example-init",
                value="pyrebar.apps.example:initialize",
                group=Plugins.PREINIT_GROUP,
            )
        ),
        post_init=(
            EntryPoint(
                name="example-init",
                value="pyrebar.apps.example:initialize",
                group=Plugins.POSTINIT_GROUP,
            )
        ),
        apps=(
            EntryPoint(
                name="example-init",
                value="pyrebar.apps.example",
                group=Plugins.APP_GROUP,
            )
        ),
        shutdown=(
            EntryPoint(
                name="example-init",
                value="pyrebar.apps.example:shutdown",
                group=Plugins.SHUTDOWN_GROUP,
            )
        ),
    )


@pytest.mark.framework
def test_add_app():
    """Test adding an app."""
    config_args = unittest.mock.MagicMock()

    parser = unittest.mock.MagicMock()

    plugin = PluginModule(
        helpstr="",
        command=lambda x: 0,
        conf=config_args,
        func=lambda x: 0,
        aliases=[],
        logger_name=None,
    )

    pyrebar.application._add_app(parser=parser, plugin=plugin)

    config_args.assert_called_with(parser)


@pytest.mark.framework
def test_add_app2():
    """Test adding an app."""
    parser = argparse.ArgumentParser()

    plugin = PluginModule(
        helpstr="",
        command=lambda x: 0,
        conf=None,
        func=None,
        aliases=[],
        logger_name=None,
    )

    pyrebar.application._add_app(parser=parser, plugin=plugin)

    args = parser.parse_args(args="")
    args.func()


@unittest.mock.patch("sys.exit")
def test_main(mock_exit):
    """Run an example application."""
    Plugins.add_entrypoint(
        EntryPoint(
            name="example-init",
            value="pyrebar.apps.example:initialize",
            group=Plugins.PREINIT_GROUP,
        )
    )
    Plugins.add_entrypoint(
        EntryPoint(
            name="example-init",
            value="pyrebar.apps.example:initialize",
            group=Plugins.POSTINIT_GROUP,
        )
    )
    Plugins.add_entrypoint(
        EntryPoint(
            name="example-init", value="pyrebar.apps.example", group=Plugins.APP_GROUP
        )
    )
    Plugins.add_entrypoint(
        EntryPoint(
            name="example-init",
            value="pyrebar.apps.example:shutdown",
            group=Plugins.SHUTDOWN_GROUP,
        )
    )
    pyrebar.application.main(argv=["--info"])



"""Load and manage application plugins."""
import inspect
import sys
from dataclasses import dataclass
from typing import Any

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points, EntryPoint, EntryPoints
else:
    from importlib.metadata import entry_points, EntryPoint, EntryPoints


@dataclass(frozen=True)
class PluginModule:
    """Plugin module data class, describing a plugin."""

    helpstr: str
    """The help string to display for this plugin."""
    command: str
    """The command to use on the command line."""
    conf: Any
    """Function to configure the command line parameters."""
    func: Any
    """The execution function for this plugin."""
    aliases: list
    """Various aliases for this module."""
    logger_name: str
    """Name of the logger to automatically configure."""


@dataclass(frozen=True)
class ProcessedPlugins:
    """Data after processing the various entrypoints."""

    pre_init: tuple[EntryPoint]
    """Pre-init plugin functions."""
    post_init: tuple[EntryPoint]
    """Post-init plugin functions."""
    apps: tuple[EntryPoint]
    """Application plugin modules."""
    shutdown: tuple[EntryPoint]
    """Post-application shutdown."""

    @staticmethod
    def loadapp(ep: EntryPoint) -> PluginModule:
        """Load an process an entrypoint into a plugin module.

        Args:
            ep (EntryPoint): The `EntryPoint` representing a plugin module.

        Returns:
            PluginModule: The processed plugin module.
        """
        module = ep.load()
        aliases = []
        helpstr = ""
        func = None
        conf = None
        command = ep.name
        logger_name = module.__name__

        for n, value in inspect.getmembers(module):
            if n == "__doc__":
                helpstr = value
            elif n == "SUBCOMMAND":
                command = value
            elif n == "config_args":
                conf = value
            elif n == "execute":
                func = value
            elif n == "ALIASES":
                aliases = value
            elif n == "LOGGER_NAME":
                logger_name = value

        return PluginModule(
            helpstr=helpstr,
            func=func,
            conf=conf,
            command=command,
            aliases=aliases,
            logger_name=logger_name,
        )


class Plugins:
    """Module enabling bootstrapping the `entry_points()` list."""

    PREINIT_GROUP = "pyrebar.preinit"
    """Group used for the pre-init step."""

    POSTINIT_GROUP = "pyrebar.postinit"
    """Group used for the post-init step."""

    APP_GROUP = "pyrebar.app"
    """Group used for applications."""

    SHUTDOWN_GROUP = "pyrebar.shutdown"
    """Group used for the post-application shutdown."""

    __entrypoints: list[EntryPoint] = []

    @classmethod
    def add_entrypoint(cls, ep: EntryPoint):
        """Explicitly add an entrypoint.

        These `EntryPoint` instances will be provided in addition to the results of those
        from the importlib metadata module.  Call this method to bootstrap applications
        running from `__main__` rather than as an installed module.

        Args:
            ep (EntryPoint): The entrypoint to add.
        """
        if ep:
            cls.__entrypoints.append(ep)

    @classmethod
    def entry_points(cls) -> EntryPoints:
        """Provide entry points from the installed modules.

        This method aggregates the installed modules with any included via
        bootstrapping.

        Returns:
            EntryPoints: The resulting entrypoints
        """
        points = entry_points()

        for e in cls.__entrypoints:
            if e.name in points.names:
                points[e.name].append(e)
            else:
                points[e.name] = [e]

        return points

    @classmethod
    def extract_plugins(cls) -> ProcessedPlugins:
        """Extract the py-rebar plugins from the entrypoints.

        Returns:
            ProcessedPlugins: The processed entrypoints
        """
        entry_points = Plugins.entry_points()

        pre_init = tuple(e for e in entry_points.select(group=Plugins.PREINIT_GROUP))
        post_init = tuple(e for e in entry_points.select(group=Plugins.POSTINIT_GROUP))
        apps = tuple(e for e in entry_points.select(group=Plugins.APP_GROUP))
        shutdown = tuple(e for e in entry_points.select(group=Plugins.SHUTDOWN_GROUP))

        return ProcessedPlugins(
            pre_init=pre_init, post_init=post_init, apps=apps, shutdown=shutdown
        )



"""Unit tests for utils/logging.py."""
import argparse
import copy
from unittest.mock import patch
import pytest

import pyrebar.utils.logging


@pytest.fixture(autouse=True)
def test_config():
    """Fixture for the logging configuration."""
    config = copy.deepcopy(pyrebar.utils.logging.CONFIG)
    yield pyrebar.utils.logging.CONFIG
    pyrebar.utils.logging.CONFIG = config


def test_config_args():
    """Verify creating the command-line arguments."""
    parser = argparse.ArgumentParser()
    pyrebar.utils.logging.config_args(parser)

    # verify the log levels
    for param, expected in (
        (None, "INFO"),
        ("--quiet", "CRITICAL"),
        ("--critical", "CRITICAL"),
        ("--error", "ERROR"),
        ("--warn", "WARNING"),
        ("--warning", "WARNING"),
        ("--info", "INFO"),
        ("--debug", "DEBUG"),
    ):
        args = parser.parse_args([param] if param else [])
        assert expected == args.log_level
        assert not args.log_root_level

    # verify the root log levels
    for param, expected in (
        ("--root-quiet", "CRITICAL"),
        ("--root-critical", "CRITICAL"),
        ("--root-error", "ERROR"),
        ("--root-warn", "WARNING"),
        ("--root-warning", "WARNING"),
        ("--root-info", "INFO"),
        ("--root-debug", "DEBUG"),
    ):
        args = parser.parse_args([param] if param else [])
        assert "INFO" == args.log_level
        assert expected == args.log_root_level


@patch("pyrebar.utils.logging.logging.config.dictConfig")
def test_config_logging_noargs(mock_logging_config, test_config):
    """Verify logging is configured without args."""
    expected = copy.deepcopy(test_config)
    expected["loggers"][""]["level"] = "ERROR"

    pyrebar.utils.logging.config_logging(None)

    mock_logging_config.assert_called_with(expected)


@patch("pyrebar.utils.logging.logging.config.dictConfig")
def test_config_logging_rootinfo(mock_logging_config, test_config):
    """Verify logging is configured with the root log level set to info."""
    expected = copy.deepcopy(test_config)
    expected["loggers"][""]["level"] = "INFO"

    pyrebar.utils.logging.config_logging(argparse.Namespace(log_root_level="INFO"))

    mock_logging_config.assert_called_with(expected)


@patch("pyrebar.utils.logging.logging.config.dictConfig")
def test_config_logging_info(mock_logging_config, test_config):
    """Verify logging is configured with the log level set to info."""
    expected = copy.deepcopy(test_config)
    expected["loggers"][""]["level"] = "INFO"

    pyrebar.utils.logging.config_logging(
        argparse.Namespace(log_level="INFO", log_root_level="ERROR")
    )

    mock_logging_config.assert_called_with(expected)


@patch("pyrebar.utils.logging.logging.config.dictConfig")
def test_config_logging_name(mock_logging_config, test_config):
    """Verify logging is configured with the logger name set."""
    expected = copy.deepcopy(test_config)
    expected["loggers"][""]["level"] = "ERROR"
    expected["loggers"]["yellowbeard"] = copy.deepcopy(expected["loggers"][""])
    expected["loggers"]["yellowbeard"]["level"] = "INFO"

    pyrebar.utils.logging.config_logging(
        argparse.Namespace(
            log_level="INFO", log_root_level="ERROR", logger_name="yellowbeard"
        )
    )

    mock_logging_config.assert_called_with(expected)


@patch("pyrebar.utils.logging.logging.config.dictConfig")
def test_config_logging_func(mock_logging_config, test_config):
    """Verify logging is configured with a logger name inferred from function."""
    expected = copy.deepcopy(test_config)
    expected["loggers"][""]["level"] = "ERROR"
    expected["loggers"]["tests.test_utils_logging"] = copy.deepcopy(
        expected["loggers"][""]
    )
    expected["loggers"]["tests.test_utils_logging"]["level"] = "INFO"

    pyrebar.utils.logging.config_logging(
        argparse.Namespace(
            log_level="INFO", log_root_level="ERROR", func=test_config_logging_func
        )
    )

    mock_logging_config.assert_called_with(expected)



"""Test appliction.

Running this script simulates a build + install + run of the module.
"""
from pyrebar.application import main
from pyrebar import Plugins
from importlib.metadata import EntryPoint
import sys

if __name__ == "__main__":
    """Bootstrap the plugins"""
    Plugins.add_entrypoint(
        EntryPoint(
            name="example-init",
            value="pyrebar.apps.example:initialize",
            group=Plugins.PREINIT_GROUP,
        )
    )
    Plugins.add_entrypoint(
        EntryPoint(
            name="example-init",
            value="pyrebar.apps.example:initialize",
            group=Plugins.POSTINIT_GROUP,
        )
    )
    Plugins.add_entrypoint(
        EntryPoint(
            name="example-init", value="pyrebar.apps.example", group=Plugins.APP_GROUP
        )
    )

    rc = main()
    sys.exit(rc)



"""Unit tests for __main__.py."""
import pytest
from unittest.mock import patch

import pyrebar.__main__ as pymain


@pytest.mark.framework
@patch.object(pymain, "__name__", "__main__")
@patch.object(pymain, "main")
@patch.object(pymain.sys, "exit")
def test_main(mock_exit, mock_main):
    """Unit test the __main__ to be sure the application is executed."""
    mock_main.return_value = 42

    pymain.run_main()

    mock_main.assert_called()
    mock_exit.assert_called_with(42)



"""Singleton metaclass."""


class SingletonMeta(type):
    """Metaclass for singleton objects."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        """Execute the metaclass."""
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
            cls.x = 5
        return cls._instances[cls]



"""Application framework."""
import argparse
import logging
import pyrebar.utils.logging
from .plugins import Plugins, ProcessedPlugins, PluginModule


def _add_app(parser: argparse.ArgumentParser, plugin: PluginModule):
    if plugin.func:
        parser.set_defaults(func=plugin.func, logger_name=plugin.logger_name)
    else:
        parser.set_defaults(func=lambda **kwargs: None, logger_name="")

    if plugin.conf:
        plugin.conf(parser)


def main(argv=None) -> int:
    """The main application.

    Returns:
        int: The return value.
    """
    plugins = Plugins.extract_plugins()
    parser = argparse.ArgumentParser()

    # add the pyrebar logging parameters
    pyrebar.utils.logging.config_args(parser)

    try:
        # pre-init
        for e in plugins.pre_init:
            e.load()(parser=parser)

        # load plugins
        if not plugins.apps:
            raise ModuleNotFoundError(
                "No apps loaded. Verify a business logic application is installed."
            )
        elif len(plugins.apps) == 1:
            plugin = ProcessedPlugins.loadapp(plugins.apps[0])
            _add_app(parser, plugin)
        else:
            subparsers = parser.add_subparsers(
                title="Subcommands",
                description="Valid subcommands.",
                help="Specify {subcommand} --help for more details",
            )

            for e in plugins.apps:
                plugin = ProcessedPlugins.loadapp(e)
                if plugin.func:
                    p = subparsers.add_parser(
                        plugin.command, help=plugin.helpstr, aliases=plugin.aliases
                    )
                    _add_app(p, plugin)

        # process arguments
        args = parser.parse_args(args=argv)

        # configure the logger
        pyrebar.utils.logging.config_logging(args)

        # post-init
        for e in plugins.post_init:
            e.load()(args=args)

        # execute the selected plugin
        if args.func:
            return args.func(args=args)
        else:
            raise RuntimeError("No subcommand specified, use --help for more info.")

    finally:
        for e in plugins.shutdown:
            try:
                e.load()()
            except:  # noqa: E722 - log shutdown errors
                logging.getLogger(__name__).error(
                    "Caught error executing shutdown hook %s",
                    e.name,
                    exc_info=1,
                    stack_info=1,
                )



"""Example application."""
import logging

SUBCOMMAND = "example-command"
"""Override the default subcommand name. Remove if unnecessary."""

ALIASES = ["exc"]
"""Add subcommand aliases."""

LOGGER_NAME = "pyrebar.apps"


def initialize(**kwargs):
    """Initialize the framework."""
    print("Initialized the pirate ship")


def config_args(parser):
    """Add example command line parameters."""
    parser.add_argument("--pillage", help="Pillage in piratey ways.")


def execute(args=None) -> int:
    """An example application.

    This function performs a basic thing that does stuff.
    """
    logger = logging.getLogger(__name__)

    print(__name__)

    print("Yay example!")

    logger.info("completed.")
    return 0


def shutdown():
    """Shutdown hook."""
    print("shutdown hook")



"""Configure the python logger."""
import logging
import logging.config


CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {
        "standard": {
            "format": "%(asctime)s %(levelname)s %(name)s %(message)s",
            "datefmt": "%Y-%m-%dT%H:%M:%S%z",
        }
    },
    "handlers": {
        "standard": {"class": "logging.StreamHandler", "formatter": "standard"}
    },
    "loggers": {
        "": {
            "handlers": ["standard"],
            "level": "ERROR",
            "propagate": False,
        }
    },
}
"""Config dictionary."""


def config_args(parser):
    """Add command line arguments for logging.

    Args:
        parser (argparse.ArgumentParser): The command line parser.
    """
    loglevel = parser.add_argument_group(
        title="Log level", description="Set detail level of application log output."
    ).add_mutually_exclusive_group()
    loglevel.add_argument(
        "--quiet",
        "--critical",
        action="store_const",
        const="CRITICAL",
        dest="log_level",
        help="Suppress all but the most critical log statements.",
    )

    loglevel.add_argument(
        "--error",
        action="store_const",
        const="ERROR",
        dest="log_level",
        help="Display error logs.",
    )
    loglevel.add_argument(
        "--warn",
        "--warning",
        action="store_const",
        const="WARNING",
        dest="log_level",
        help="Display error and warning logs.",
    )
    loglevel.add_argument(
        "--info",
        action="store_const",
        const="INFO",
        dest="log_level",
        help="Print informational logging.",
    )
    loglevel.add_argument(
        "--debug",
        action="store_const",
        const="DEBUG",
        dest="log_level",
        help="Display highly detailed level of logging.",
    )

    loglevel = parser.add_argument_group(
        title="Root log level", description="Set detail level of root log output."
    ).add_mutually_exclusive_group()
    loglevel.add_argument(
        "--root-quiet",
        "--root-critical",
        action="store_const",
        const="CRITICAL",
        dest="log_root_level",
        help="Suppress all but the most critical log statement.",
    )

    loglevel.add_argument(
        "--root-error",
        action="store_const",
        const="ERROR",
        dest="log_root_level",
        help="Display error logs.",
    )
    loglevel.add_argument(
        "--root-warn",
        "--root-warning",
        action="store_const",
        const="WARNING",
        dest="log_root_level",
        help="Display error and warning logs.",
    )
    loglevel.add_argument(
        "--root-info",
        action="store_const",
        const="INFO",
        dest="log_root_level",
        help="Print informational logging.",
    )
    loglevel.add_argument(
        "--root-debug",
        action="store_const",
        const="DEBUG",
        dest="log_root_level",
        help="Display highly detailed level of logging.",
    )
    parser.set_defaults(log_root_level=None, log_level="INFO")


def config_logging(args=None):
    """Configure logging based on the provided command line arguments.

    Args:
        args (argparse.Namespace, optional): The parsed command line arguments.
        Defaults to None.
    """
    app_logger_name = None
    if args:
        if "logger_name" in args and args.logger_name is not None:
            app_logger_name = args.logger_name
        elif "func" in args and args.func:
            m: str = args.func.__module__
            f: str = args.func.__name__
            app_logger_name = m.removesuffix(f".{f}")

    root_level = _arg_or_default(args, "log_root_level", "ERROR")

    if app_logger_name:
        CONFIG["loggers"][app_logger_name] = {
            "handlers": ["standard"],
            "level": _arg_or_default(args, "log_level", "INFO"),
            "propagate": False,
        }
    else:
        root_level = _arg_or_default(args, "log_level", root_level)

    CONFIG["loggers"][""]["level"] = root_level

    logging.config.dictConfig(CONFIG)


def _arg_or_default(args, key: str, default=None):
    if args and key in args:
        v = vars(args)[key]
        return v or default
    else:
        return default



"""Unit tests for utils/singleton_meta.py."""
from pyrebar.utils import SingletonMeta


class TestClass(metaclass=SingletonMeta):
    """Test singleton class."""

    value: str = "foo"


def test_singleton():
    """Unit test for the SingletonMeta metaclass."""
    a = TestClass()
    b = TestClass()

    assert a.value == b.value
    assert a == b

    a.value = "bar"
    assert "bar" == b.value



"""Example application."""
import logging


def execute(args=None) -> int:
    """An example application.

    This function performs a basic thing that does stuff.
    """
    logger = logging.getLogger(__name__)

    print("Flee!! Yellowbeard cometh!")

    logger.info("completed.")
    return 0
